from __future__ import annotations
from collections import defaultdict
########################################################################################################################

__doc__ = \
    """
Combined RMSD
    """

__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "2020 A.D."
__license__ = "MIT"
__citation__ = ""

import warnings

from .version import __version__

########################################################################################################################


from rdkit import Chem
from rdkit.Chem import rdFMCS, AllChem
from rdkit.Geometry import Point3D

from typing import Sequence, List, Optional, Tuple

import json, re
import numpy as np

from .monster import Monster
from typing import *

class mRMSD:
    """Mols are non-superposed ('aligned') for the RMSD and in Å.

    The RMSD has been calculated differently.
    The inbuilt RMSD calculations in RDKit (``Chem.rdMolAlign.GetBestRMS``) align the two molecules,
    this does not align/superpose them.
    This deals with the case of multiple hits.
    As a comparision, For euclidean distance the square root of the sum of the differences in each coordinates is taken.
    As a comparision, For a regular RMSD the still-squared distance is averaged before taking the root.
    Here the average is done across all the atom pairs between each hit and the followup.
    Therefore, atoms in followup that derive in the blended molecule by multiple atom are scored multiple times.

    .. math::

        \sqrt{\frac{\sum_{i}^{N_{\rm{hits}}} (\sum_{i}^{n} (q_{i,\rm{x}} - h_{i,\rm{x}})^2 +  \\
        (q_{i,\rm{y}} - h_{i,\rm{y}})^2 + (q_{i,\rm{z}} - h_{i,\rm{z}})^2 }{n\cdot m}}



    """
    # faux enum:
    HIT_BASED = 0  # regularly instantiated
    XYZ_BASED = 1  # from_internal_xyz
    IDENTITY_BASED = 2  # identity

    def __init__(self,
                 followup: Chem.Mol,
                 hits: Sequence[Chem.Mol],
                 mappings: List[List[Tuple[int, int]]]):
        """
        This is not meant to be called directly.
        mappings is a list of len(hits) containing lists of tuples of atom idx that go from followup to hit


        The hit _Name must match that in origin!
        currected output of monster.origin_from_mol() or cls.get_origins(to-be-scored-mol, annotated)

        :param followup: the followup compounds
        :param hits: the fragment hits
        :param mappings: a complicated affair...
        """
        self.followup = followup
        self.hits = hits
        self.mappings = mappings
        self.mode = self.HIT_BASED
        # calculate
        mds = []
        self.rmsds = []
        tatoms = 0
        for hit, mapping in zip(hits, mappings):
            md = self.calculate_msd(self.followup, hit, mapping)
            mds.append(md)
            tatoms += len(mapping)
            if len(mapping):
                self.rmsds.append((md/ len(mapping)) ** 0.5)
            else:
                self.rmsds.append(None)
        if tatoms:
            self.mrmsd = (sum(mds) / tatoms) ** 0.5
        else:
            self.mrmsd = None

    @staticmethod
    def generate_overlap_mapping(mol_a: Chem.Mol, mol_b: Chem.Mol) -> List[Tuple[int, int]]:
        # Monster.get_positional_mapping : Dict[int, int]
        map_dict : Dict[int, int] = Monster.get_positional_mapping(mol_a, mol_b)
        return list(map_dict.items())

    @classmethod
    def from_unannotated_mols(cls,
                            moved_followup: Chem.Mol,
                            hits: Sequence[Chem.Mol],
                            placed_followup: Chem.Mol
                            ) -> mRMSD:
        """
        Mapping is done by positional overlap between placed_followup and hits
        This mapping is the applied to moved_followup.
        The mapping is not between placed and moved. But the former acts as a go between.

        :param moved_followup: The mol to be scored
        :param hits: the hits to score against
        :param placed_followup: the mol to determine how to score
        :return:
        """
        placed2moved : Dict[int, int] = dict(cls.from_unannotated_same_mols(placed_followup=placed_followup,
                                                                            moved_followup=moved_followup
                                                                            ).mappings[0]
                                             )
        mappings = []
        for h, hit in enumerate(hits):
            placed2hit : List[Tuple[int, int]] = cls.generate_overlap_mapping(placed_followup, hit)
            mappings.append([(placed2moved[pi], hi) for pi, hi in placed2hit if pi in placed2moved])
        return cls(followup=moved_followup, hits=hits, mappings=mappings)

    @classmethod
    def from_unannotated_same_mols(cls,
                            moved_followup: Chem.Mol,
                            placed_followup: Chem.Mol
                            ) -> mRMSD:
        """
        Mapping is not done by positional overlap between placed_followup and moved_followup
        But by shape overlap that yields the lowest RMSD
        This means that every atom matches reguardless of hits.

        :param moved_followup: The mol to be scored
        :param placed_followup: the mol to determine how to score
        :return:
        """
        # placed idx -> moved idx
        munge_mapping = lambda match: list(zip(range(placed_followup.GetNumAtoms()), match))
        candidate_mappings = [munge_mapping(match) for match in moved_followup.GetSubstructMatches(placed_followup)]
        assert candidate_mappings, 'No matches?!'
        # mappings is places to moved hence the order
        candidate_mrmsds = [cls(followup=placed_followup, hits=[moved_followup], mappings=[mapping])
                            for mapping in candidate_mappings]
        self = sorted(candidate_mrmsds, key=lambda m: m.mrmsd)[0]
        # self.hits is placed_followup not hits
        self.mode = self.IDENTITY_BASED
        return self

    @classmethod
    def from_annotated_mols(cls,
                  annotated_followup: Chem.Mol,
                  hits: Optional[Sequence[Chem.Mol]]=None
                  ) -> mRMSD:
        """
        Monster leaves a note of what it did. atom prop _Origin is a json of a list of mol _Name dot AtomIdx.
        This classmethod accepts a followup with has this.

        :param annotated_followup:
        :param hits:
        :return:
        """
        if cls.is_xyz_annotated(annotated_followup):
            self = cls.from_internal_xyz(annotated_followup)
            self.hits = hits
            return self
        mappings = cls._mapping_from_annotated_and_hits(annotated_followup, hits)
        return cls(annotated_followup, hits, mappings)

    @classmethod
    def is_origin_annotated(cls, mol: Chem.Mol) -> bool:
        """
        This is atom not mol.
        """
        for atom in mol.GetAtoms():
            if len(cls._get_origin(atom)) > 0:
                return True
        else:
            return False

    @classmethod
    def is_xyz_annotated(cls, mol: Chem.Mol) -> bool:
        for atom in mol.GetAtoms():
            if len(cls._get_xyz(atom)) > 0:
                return True
        else:
            return False

    @classmethod
    def _mapping_from_annotated_and_hits(cls,
                                         annotated_followup: Chem.Mol,
                                         hits: Sequence[Chem.Mol]):
        assert cls.is_origin_annotated(annotated_followup), 'This molecules is not annotated.'
        mappings = []
        for h, hit in enumerate(hits):
            hname = hit.GetProp('_Name')
            mapping = []
            if hname == '':
                print(f'{hit} has no name!')
            else:
                for i in range(annotated_followup.GetNumAtoms()):
                    atom = annotated_followup.GetAtomWithIdx(i)
                    for oel in cls._get_origin(atom):
                        rex = re.match(hname + '\.(\d+)', oel)
                        if rex is not None:
                            h = int(rex.group(1))
                            mapping.append((i, h))
            mappings.append(mapping)
        return mappings

    @classmethod
    def from_other_annotated_mols(cls,
                            followup: Chem.Mol,
                            hits: Sequence[Chem.Mol],
                            annotated: Chem.Mol,
                            ) -> mRMSD:
        """
        The two molecules are the same (atom names included) but the followup lacks annotations.
        """
        cls.overannotate(annotated, hits=hits, priority=('atom_origin', 'mol_origin', 'xyz'))
        if all([cls.is_xyz_annotated(annotated),
                annotated.GetAtomWithIdx(0).GetPDBResidueInfo() is not None,
                followup.GetAtomWithIdx(0).GetPDBResidueInfo() is not None
                ]):
            # match by atom name
            name2xyz = {}
            for atom in annotated.GetAtoms():
                pi = atom.GetPDBResidueInfo()
                if pi and atom.HasProp('_x'):
                    name2xyz[pi.GetName()] = {cart: atom.GetDoubleProp("_" + cart) for cart in ('x', 'y', 'z')}
            for atom in followup.GetAtoms():
                pi = atom.GetPDBResidueInfo()
                if pi is None:
                    continue
                name = pi.GetName()
                if pi and name in name2xyz:
                    for cart in ('x', 'y', 'z'):
                        atom.SetDoubleProp('_' + cart, name2xyz[name][cart])
            return cls.from_annotated_mols(followup, hits)
        elif (atom.GetSymbol() for atom in annotated.GetAtoms()) == (atom.GetSymbol() for atom in followup.GetAtoms()):
            # element sequence matches... but has no atom names
            for anno_atom, follow_atom in zip(annotated.GetAtoms(), followup.GetAtoms()):
                if anno_atom.HasProp('_Origin'):
                    follow_atom.SetProp('_Origin', anno_atom.GetProp('_Origin'))
                if anno_atom.HasProp('_x'):
                    follow_atom.SetProp('_x', anno_atom.GetProp('_x'))
                    follow_atom.SetProp('_y', anno_atom.GetProp('_y'))
                    follow_atom.SetProp('_z', anno_atom.GetProp('_z'))
        else:
            # has become nearly redundant with from_unannotated_mols except this expect annotated to have:
            # from_unannotated_mols
            # the former way has issues with isomorphisms
            # cls.copy_origins(annotated, followup)
            # return cls.from_annotated_mols(followup, hits)
            posibilities : Tuple[List[Chem.Mol], List[List[int]]]= cls.copy_all_possible_origins(annotated, followup)
            targets: List[Chem.Mol] = posibilities[0]
            assert targets, 'Molecule could not be mapped.'
            results = [(target, cls.from_annotated_mols(target, hits)) for target in targets]
            return list(sorted(results, key=lambda x: x[1].mrmsd))[0][1]


    def calculate_msd(self, molA, molB, mapping) -> float:
        """
        A nonroot rmsd.

        :param molA:
        :param molB:
        :param mapping: lists of tuples of atom idx that go from molA to molB
        :return: nonroot rmsd
        """
        confA = molA.GetConformer()
        confB = molB.GetConformer()
        return sum([(confA.GetAtomPosition(a).x - confB.GetAtomPosition(b).x) ** 2 +
                    (confA.GetAtomPosition(a).y - confB.GetAtomPosition(b).y) ** 2 +
                    (confA.GetAtomPosition(a).z - confB.GetAtomPosition(b).z) ** 2 for a, b in mapping])

    def calculate_rmsd(self, molA, molB, mapping) -> float:
        return (self.calculate_msd(molA, molB, mapping) / len(mapping)) ** 0.5

    @classmethod
    def mock(cls):
        self = cls.__new__(cls)
        self.followup = None
        self.hits = None
        self.mappings = {}
        self.mrmsd = float('nan')
        self.rmsds = []
        return self


    @classmethod
    def copy_origins(cls, annotated: Chem.Mol, target: Chem.Mol):
        """
        > This is no longer used by ``from_other_annotated_mols``.

        Monster leaves a note of what it did. atom prop _Origin is a json of a list of mol _Name dot AtomIdx.
        However, the atom order seems to be maintained but I dont trust it. Also dummy atoms are stripped.

        :param annotated:
        :param target:
        :return: a list of origins
        """
        warnings.warn('This is no longer used by ``from_other_annotated_mols``.', DeprecationWarning)
        mcs = rdFMCS.FindMCS([target, annotated],
                             atomCompare=rdFMCS.AtomCompare.CompareElements,
                             bondCompare=rdFMCS.BondCompare.CompareAny,
                             ringMatchesRingOnly=True)
        common = Chem.MolFromSmarts(mcs.smartsString)
        dmapping = dict(zip(target.GetSubstructMatch(common), annotated.GetSubstructMatch(common)))
        origins = []
        cls.overannotate(annotated, [], ('atom_origin', 'mol_origin', 'surrender'))
        for i in range(target.GetNumAtoms()):
            if i in dmapping:
                atom = annotated.GetAtomWithIdx(dmapping[i])
                tatom = target.GetAtomWithIdx(i)
                o = cls._get_origin(atom)
                tatom.SetProp('_Origin', json.dumps(o))
        return origins

    @classmethod
    def copy_all_possible_origins(cls, annotated: Chem.Mol, target: Chem.Mol) -> Tuple[List[Chem.Mol], List[List[int]]]:
        """
        Monster leaves a note of what it did. atom prop _Origin is a json of a list of mol _Name dot AtomIdx.
        However, the atom order seems to be maintained but I dont trust it. Also dummy atoms are stripped.

        :param annotated:
        :param target:
        :return: a list of mols and a list of orgins (a list too)
        """
        cls.overannotate(annotated, [], ('atom_origin', 'mol_origin', 'xyz'))
        mcs = rdFMCS.FindMCS([target, annotated],
                             atomCompare=rdFMCS.AtomCompare.CompareElements,
                             bondCompare=rdFMCS.BondCompare.CompareAny,
                             ringMatchesRingOnly=True)
        common = Chem.MolFromSmarts(mcs.smartsString)
        options = []
        originss = []
        # adding Unique=False will make benzene isomorphism not an issue, but the code will grind to a halt.
        for target_match in target.GetSubstructMatches(common):
            for anno_match in annotated.GetSubstructMatches(common):
                dmapping = dict(zip(target_match, anno_match))
                origins = []
                option = Chem.Mol(target)
                for i in range(option.GetNumAtoms()):
                    if i in dmapping:
                        atom = annotated.GetAtomWithIdx(dmapping[i])
                        tatom = option.GetAtomWithIdx(i)
                        o = cls._get_origin(atom)
                        tatom.SetProp('_Origin', json.dumps(o))
                        xyz = cls._get_xyz(atom)
                        if xyz:
                            cls._set_xyz(tatom, xyz)
                options.append(option)
                originss.append(origins)
        return options, originss

    @classmethod
    def overannotate(cls, mol: Chem.Mol, hits: List[Chem.Mol], priority=('mol_origin', 'atom_origin', 'xyz')):
        """
        Unfortunately, in an attempt to make users happy, I have to make the code more complicated.
        There are three different annotation systems going on.
        The first is the property '_Origin' in the ``Chem.Mol``.
        The second is the property '_Origin' in the ``Chem.Atom``.
        The third is the properties '_x', '_y', '_z' in the ``Chem.Atom``.

        This method will take the first priority that is available and propagate it to the other two.

        :param mol: the annotated molecule
        :param hits: the hits that are used to annotate the molecule
        :param priority: the priority of the annotation, a sequence of up to four strings:
            'mol_origin', 'atom_origin', 'xyz', 'surrender'
        """
        if len(priority) == 0:
            raise ValueError('molecule is not annotated')
        task = priority[0]
        # -----------------------------------------------
        if task == 'mol_origin':
            if not mol.HasProp('_Origin'):
                return cls.overannotate(mol, hits, tuple(priority[1:]))
            origins = json.loads(mol.GetProp('_Origin'))
            assert len(origins) == mol.GetNumAtoms(), f'Mismatch {len(origins)} vs. {mol.GetNumAtoms()}'
            for i, atom in enumerate(mol.GetAtoms()):
                # add origin to atom
                atom.SetProp('_Origin', json.dumps(origins[i]))
                # add xyz to atom
                cls._atom_origin_to_xyz(origins[i], atom, hits)
            return None
        # -----------------------------------------------
        elif task == 'atom_origin':
            if not cls.is_origin_annotated(mol):
                return cls.overannotate(mol, hits, tuple(priority[1:]))
            origins = [cls._get_origin(atom) for atom in mol.GetAtoms()]
            # add origin to mol
            mol.SetProp('_Origin', json.dumps(origins))
            # add xyz to atom
            for atom, origin in zip(mol.GetAtoms(), origins):
                cls._atom_origin_to_xyz(origin, atom, hits)
        # -----------------------------------------------
        elif task == 'xyz':
            if not cls.is_xyz_annotated(mol):
                return cls.overannotate(mol, hits, tuple(priority[1:]))
            # this cannot be consolidated...
            pass
        elif task == 'surrender':
            return None
        else:
            raise ValueError(f'Unknown task {task}')

    @classmethod
    def _atom_origin_to_xyz(cls, origins: List[str], atom: Chem.Atom, hits: List[Chem.Mol]):
        points = []
        if len(hits) == 0:
            return  # no hits
        hitdex: Dict[str, Chem.Mol] = {hit.GetProp('_Name'): hit for hit in hits}
        for origin in origins:
            if origin == 'none':
                # thank you human.
                continue
            hit_name, idx = re.match(r'(.*)\.(\d+)', origin).groups()
            atom.SetProp('_ori_name', hit_name)
            if hit_name not in hitdex:
                return
            ref_conf = hitdex[hit_name].GetConformer()
            ref_xyz: Point3D = ref_conf.GetAtomPosition(int(idx))
            points.append(list(ref_xyz))  # noqa this is okay
        if points:
            xyz = np.mean(points, axis=0)
            cls._set_xyz(atom, xyz)

    @classmethod
    def migrate_origin(cls, mol: Chem.Mol, tag='_Origin') -> Chem.Mol:
        """
        The origin list may be saved as a molecule property rather than an atom -saved as a mol say.

        :param mol: mol to fix
        :param tag: name of prop
        :return: the same mol
        """
        raise NotImplementedError('This method was deprecated. Use ``overannotate`` instead. Requires HITS!')

    @classmethod
    def _get_origin(cls, atom: Chem.Atom) -> List[str]:
        if atom.HasProp('_Origin'):
            o = atom.GetProp('_Origin')
            if o != 'none':
                return json.loads(o)
            else:
                return []
        else:
            return []

    @classmethod
    def _get_xyz(cls, atom: Chem.Atom) -> Tuple[float,float,float]:
        if atom.HasProp('_x'):
            return (atom.GetDoubleProp('_x'),
                    atom.GetDoubleProp('_y'),
                    atom.GetDoubleProp('_z'))
        else:
            return ()

    @classmethod
    def _set_xyz(cls, atom: Chem.Atom, xyz):
        if len(xyz):
            atom.SetDoubleProp('_x', float(xyz[0])),
            atom.SetDoubleProp('_y', float(xyz[1])),
            atom.SetDoubleProp('_z', float(xyz[2]))

    @classmethod
    def from_internal_xyz(cls, annotated_followup):
        """
        This is an alternative for when the atoms have _x, _y, _z
        
        :param annotated_followup:
        :return:
        """
        self = cls.__new__(cls)
        self.followup = annotated_followup
        self.hits = []
        self.mappings = []
        self.rmsds = []
        self.mrmsd = float('nan')
        self.mode = self.XYZ_BASED
        conf = annotated_followup.GetConformer()
        n = 0
        tatoms = 0
        ns = defaultdict(int)
        ts = defaultdict(int)
        for a, atom in enumerate(annotated_followup.GetAtoms()):
            if not atom.HasProp('_x'):
                continue
            x, y, z = cls._get_xyz(atom)
            tatoms += 1
            d = sum([(conf.GetAtomPosition(a).x - x) ** 2 +
                        (conf.GetAtomPosition(a).y - y) ** 2 +
                        (conf.GetAtomPosition(a).z - z) ** 2])
            n += d
            if atom.HasProp('_Origin'):
                origins = json.loads(atom.GetProp('_Origin'))
                for origin in origins:
                    if origin == 'none':
                        continue
                    hit_name, idx = re.match(r'(.*)\.(\d+)', origin).groups()
                    ns[hit_name] += d
                    ts[hit_name] += 1
            elif atom.HasProp('_ori_name'):
                hit_name = atom.GetProp('_ori_name')
                ns[hit_name] += d
                ts[hit_name] += 1
        self.mrmsd = (n / tatoms) ** 0.5
        self.rmsds = [(ns[hit_name] / ts[hit_name]) ** 0.5 for hit_name in ns]
        return self





import re, string, functools
from typing import Optional, Set

from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit_to_params import Params, Constraints
from ._victor_journal import _VictorJournal
from .minimalPDB import MinimalPDBParser
from ..monster._ff import MinizationOutcome


class _VictorPlonk(_VictorJournal):
    # the following is here as opposed to Monster, because it requires the template, ligand connections etc.
    # the stupid name "plonk" is to distinguish it from place and placement, which have a different meaning in Monster.

    def _get_LINK_record(self):
        if self.is_covalent:
            # get correct chain names.
            l_resi, l_chain = re.match('(\d+)(\D?)', str(self.ligand_resi)).groups()
            if self.covalent_resi:
                p_resi, p_chain = re.match('(\d+)(\D?)', str(self.covalent_resi)).groups()
            else:
                p_resi, p_chain = None, None
            if not p_chain:
                p_chain = 'A'
            if not l_chain:
                l_chain = 'B'
            # get the cx atom name
            assert len(self.params.CONNECT) != 0, 'No connection atom found in params file.'
            cx = self.params.pad_name(self.params.CONNECT[0].atom_name)
            # TODO the SG connection is hardcoded.
            return f'LINK         SG  {self.covalent_resn} {p_chain} {p_resi: >3}                ' + \
                f'{cx} {self.ligand_resn} {l_chain} {l_resi: >3}     1555   1555  1.8\n'
        else:
            return ''

    def _correct_ligand_info(self, mol: Optional[Chem.Mol] = None) -> Chem.Mol:
        """
        Corrects in place the given mol based on self.ligand_resi.
        If none provided it assumed self.monster.positioned_mol
        Correcting the serial unfortunately does not do anything.
        """
        i = 0
        if mol is None:
            mol = self.monster.positioned_mol
        l_resi, l_chain = re.match('(\d+)(\D?)', str(self.ligand_resi)).groups()  # TODO improve ligand_resi
        for atom in mol.GetAtoms():
            info = atom.GetPDBResidueInfo()
            if info is None:
                i += 1
                self.journal.warning(f'The atom #{atom.GetIdx()} has no PDB information in RDKit')
                info = Chem.AtomPDBResidueInfo(atomName=f'X{i: <3}')
            info.SetResidueNumber(int(l_resi))
            info.SetChainId(l_chain)
            info.SetIsHeteroAtom(True)
            info.SetOccupancy(1.)
            info.SetResidueName(self.ligand_resn)
        return mol

    def _plonk_monster_in_structure(self, prepped_mol: Optional[Chem.Mol]=None, use_pymol=False):
        prepped_mol = self._correct_ligand_info(prepped_mol)
        self.journal.debug(f'{self.long_name} - placing monster in structure')
        if use_pymol:
            return self._plonk_monster_in_structure_pymol()
        else:
            # _plonk_monster_in_structure_raw does no corrections.
            return self._plonk_monster_in_structure_minimal(prepped_mol)

    @functools.cached_property
    def preminimized_mol(self) -> Chem.Mol:
        """
        This cached property is the preminimised molecule.
        It extracts the neighbourhood (``monster.get_neighborhood``) and
        minimises the molecule (``monster.mmff_minimize``)

        This method is called by the plonking into structure methods.
        Not "positioning" as intended by ``monster`` is done.
        """
        mol = Chem.Mol(self.monster.positioned_mol)
        if self.monster_mmff_minisation:
            self.journal.debug(f'{self.long_name} - pre-minimising monster (MMFF)')
            if self.settings.get('ff_use_neighborhood', True):
                neighborhood = self.monster.get_neighborhood(self.apo_pdbblock, cutoff=self.settings['ff_neighborhood'], addHs=True)
            else:
                neighborhood = None
            # ff_max_displacement = float('nan') for fixed mode
            min_result: MinizationOutcome = self.monster.mmff_minimize(mol,
                                                    neighborhood=neighborhood,
                                                    ff_max_displacement=float(self.settings.get('ff_max_displacement', 0.)),
                                                    ff_constraint=int(self.settings.get('ff_constraint', 10)),
                                                    allow_lax=True)
            return min_result.mol


    @functools.cached_property
    def preminimized_undummied_mol(self) -> Chem.Mol:
        """
        See ``preminimized_mol``. This strips the dummy atoms from the preminimised molecule.
        """
        return AllChem.DeleteSubstructs(self.preminimized_mol, Chem.MolFromSmiles('*'))

    def _plonk_monster_in_structure_minimal(self, prepped_mol: Optional[Chem.Mol]=None) -> str:
        """
        Plonks the molecule in the structure without using pymol.
        Uses a custom miniparser. see minimalPDB.MinimalPDBParser

        :return:
        """
        # ----- load
        mol = self.preminimized_undummied_mol if prepped_mol is None else AllChem.DeleteSubstructs(prepped_mol, Chem.MolFromSmiles('*'))
        pdbdata = MinimalPDBParser(self.apo_pdbblock, remove_other_hetatms=self.remove_other_hetatms,
                                   ligname=self.ligand_resn)
        moldata = MinimalPDBParser(Chem.MolToPDBBlock(mol))
        # ------- covalent fix
        if self.is_covalent:
            pdbdata.headers.append(self._get_LINK_record())
        # ------- assertions
        l_resi, l_chain = re.match('(\d+)(\D?)', str(self.ligand_resi)).groups()  # TODO improve ligand_resi
        if pdbdata.has_residue_index(index=int(l_resi), chain=l_chain):
            raise ValueError(f'Residue {self.ligand_resi} already exists in structure')
        elif pdbdata.has_residue_name(self.ligand_resn):
            raise ValueError(f'Residue {self.ligand_resn} already exists in structure')
        # -------- append
        pdbdata.append(moldata)  # fixes offsets in ATOM/HETATM and CONECT lines.
        return str(pdbdata)

    def _correct_covalent_resi(self):
        """
        An unresolved issue is that covalent_resi acts both as a covalent residue and the reference residue.
        This corrects for the case there is no covalent_resi
        """
        pdbdata = MinimalPDBParser(self.apo_pdbblock,
                                   remove_other_hetatms=self.remove_other_hetatms,
                                   ligname=self.ligand_resn)
        entry = pdbdata.coordinates[0]
        if self.covalent_resi is None:
            self.covalent_resi = f'{pdbdata.get_residue_index(entry)}{pdbdata.get_chain(entry)}'
        else:
            p_resi, p_chain = re.match('(\d+)(\D?)', str(self.covalent_resi)).groups()
            if not pdbdata.has_residue_index(int(p_resi), p_chain):
                self.covalent_resi = f'{pdbdata.get_residue_index(entry)}{pdbdata.get_chain(entry)}'

    def _get_empty_resi(self) -> str:
        """
        return the first empty chain basically.
        """
        pdbdata = MinimalPDBParser(self.apo_pdbblock,
                                   remove_other_hetatms=False)
        chains: Set[str] = {pdbdata.get_chain(entry) for entry in pdbdata.coordinates}
        missing = sorted(set(string.ascii_uppercase).difference(chains))
        return f'1{missing[0]}'

    def _plonk_monster_in_structure_raw(self):
        """
        Plonks the molecule in the structure without using pymol.
        Not "positioning" as intended by ``monster`` is done.
        Opening a PDB in RDKit is doable but gets exponentially slow with chain length

        :return:
        """
        mol = self.preminimized_undummied_mol
        mol_block = Chem.MolToPDBBlock(mol)
        return '\n'.join([self._get_LINK_record().strip(),
                          self.apo_pdbblock.strip(),
                          mol_block
                          ]).strip()

    def _plonk_monster_in_structure_pymol(self):
        """
        Plonks the molecule in the structure using pymol.
        Not "positioning" as intended by ``monster`` is done.

        :return:
        """
        import pymol2
        mol = self.preminimized_undummied_mol
        with pymol2.PyMOL() as pymol:
            pymol.cmd.read_pdbstr(self.apo_pdbblock, 'apo')
            pos_mol = Chem.MolToPDBBlock(mol)
            # pymol.cmd.read_pdbstr(pos_mol, 'scaffold')
            pymol.cmd.remove('name R')  # no dummy atoms!
            for c in self._connected_names:
                pymol.cmd.remove(f'name {c}')  # no conns
            pymol.cmd.remove('resn UNL')  # no unmatched stuff.
            pdbblock = pymol.cmd.get_pdbstr('*')
            pymol.cmd.delete('*')
        return self._get_LINK_record() + pdbblock



import argparse
import pandas as pd
from .base import FragmensteinParserBase, _add_common, set_verbose
from rdkit import Chem
from rdkit.Chem import PandasTools
from ..laboratory import Laboratory
from ..igor import Igor
from typing import List, Tuple

class FragmensteinParserLaboratory:

    def common(self, parser: argparse.ArgumentParser):
        # hits in _add_common is for individual mol files
        _add_common(parser, hits=False, output=True, template=True)
        parser.add_argument('-i', '--input', required=True, help='input sdf file')
        parser.add_argument('-d', '--out-table', default='output.csv', help='table output file')
        parser.add_argument('-s', '--sdf-outfile', default='output.sdf', help='sdf output file')
        parser.add_argument('-c', '--cores', default=1, type=int, help='number of cores to use')
        parser.add_argument('-p', '--run-plip', default=False, type=bool, help='Run PLIP?')
        parser.add_argument('--victor',
                            help='Which victor to use: Victor, OpenVictor or Wictor',
                            default='Victor')

    def _define_laboratory(self, parser: argparse.ArgumentParser):
        """fragmenstein laboratory combine -i hits.sdf -o out.sdf
           fragmenstein laboratory place -i hits.sdf -t table.csv -o out.sdf"""
        subparsers = parser.add_subparsers(title='operation', help='combine or place')
        # combine
        combine_parser = subparsers.add_parser('combine', help='combine')
        self.common(combine_parser)
        combine_parser.set_defaults(func=self.lab_combine)
        # place
        place_parser = subparsers.add_parser('place', help='place')
        self.common(place_parser)
        place_parser.add_argument('-f', '--in-table', required=True,
                                  help='CSV table input file, ' +\
                                       'requires `name`, `smiles` and space-separated-`hit_names`')
        place_parser.set_defaults(func=self.lab_place)

    def gather(self, args: argparse.Namespace) -> Tuple[Laboratory, List[Chem.Mol]]:
        set_verbose(args.verbose)
        Igor.init_pyrosetta()
        # victor or wictor? copypasted from pipeline....
        choice = args.victor.lower()
        if choice == 'victor':
            from .victor import Victor
            Laboratory.Victor = Victor
        elif choice == 'openvictor':
            from ..openmm.openvictor import OpenVictor
            Laboratory.Victor = OpenVictor
        elif choice == 'wictor':
            from ..faux_victors import Wictor
            Laboratory.Victor = Wictor
        elif choice == 'quicktor':
            from ..faux_victors import Quicktor
            Laboratory.Victor = Quicktor
        else:
            raise ValueError(f'Unknown victor: {choice}')
        # laboratory
        lab = Laboratory(pdbblock=open(args.template).read(), run_plip=bool(args.run_plip))
        with Chem.SDMolSupplier(args.input) as suppl:
            mols = list(suppl)
        return lab, mols

    def lab_combine(self, args: argparse.Namespace) -> str:
        lab, mols = self.gather(args)
        combos: pd.DataFrame = lab.combine(mols, n_cores=args.cores)
        self.write(args, combos)

    def write(self, args: argparse.Namespace, df: pd.DataFrame):
        props = ['smiles', 'error', '∆∆G', '∆G_bound', '∆G_unbound', 'comRMSD',
                 'N_constrained_atoms', 'N_unconstrained_atoms', 'runtime', 'regarded',
                 'disregarded', 'LE', 'outcome',
                 'percent_hybrid']
        df[props].to_csv(args.out_table)
        PandasTools.WriteSDF(df, args.sdf_outfile,
                             'minimized_mol',
                             properties=props)

    def fix_columns(self, df: pd.DataFrame, preferred_name: str, options: List[str]):
        if preferred_name not in df.columns:
            for name_col in options:
                for variant_fx in [str.lower, str.upper, str.title]:
                    alt = variant_fx(name_col)  # noqa
                    if alt in df.columns:
                        df.rename(columns={alt: preferred_name}, inplace=True)
                        break
            else:
                raise ValueError(f'No name column `{preferred_name}` (or the fallbacks {options}) in {df.columns}')

    def lab_place(self, args: argparse.Namespace) -> str:
        lab, hits = self.gather(args)
        hitdex = {hit.GetProp('_Name'): hit for hit in hits}
        # fix the input table
        intable = pd.read_csv(args.in_table)
        # name
        self.fix_columns(intable, 'name', ['name', 'long name'])
        self.fix_columns(intable, 'smiles', ['smiles', 'smile'])
        self.fix_columns(intable, 'hit_names', ['hit_names', 'hit names', 'hits'])
        try:
            intable['hits'] = intable['hit_names'].apply(lambda x: [hitdex[n] for n in x.split()])
        except KeyError as error:
            raise KeyError(f'Could not find {error} in the input sdf file ({hitdex.keys()})')
        # run
        placements = lab.place(intable, n_cores=args.cores)
        self.write(args, placements)






from ..version import __version__

__doc__ = """
Command line interface to Fragmenstein.
The the first argument to the command is one of the three option:
## extract
Given a target folder (`-i <folder_name>`) 
extract the mol files (with covalent atom if present) in an output folder (`-o <folder_name>`) 
## merge
## position
NB. Please no spaces in the filenames. If spaces are present quote/double-quote the fle path.
""".strip()

from typing import *
import argparse


# from fragmenstein import Victor

class FragmensteinParser:
    ___doc__ = __doc__  # this works (it is in a function it doesn't as the global cmd is called for)

    # ===================== Actions ====================================================================================

    def extract(self, input_folder: str, output_folder: str):
        pass

    def merge(self, target_filenames: List):
        pass

    def place(self, target_filenames: List, smiles: str):
        pass

    def set_verbosity(verbose: int):
        # set logging stdout
        # logging.ERROR, logging.CRITICAL --> 0 ?
        # logging.WARNING --> 1 ?
        # logging.INFO --> 2
        # logging.DEBUG --> 3
        pass

    # ===================== Arguments ==================================================================================

    def __init__(self, cli_override: Optional[List[str]] = None):
        self.parser = self.define_parser()
        # ---- parse ---------------------------------------------------------------------------------------------------
        if cli_override:  # testing basically
            self.args = self.parser.parse_args(cli_override)
        else:  # normal way
            self.args = self.parser.parse_args()
        # ---- verify --------------------------------------------------------------------------------------------------
        self.verify()
        # ---- route ---------------------------------------------------------------------------------------------------
        self.route()

    @classmethod
    def mock(cls):
        self = cls.__new__(cls)
        self.parser = self.define_parser()
        self.args = property(lambda: exec("raise NotImplementedError"))

    def define_parser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument('mode', type=str, help='Three modes are accepted: extract, merge, position',
                            choices=['extract', 'merge', 'position'])
        parser.add_argument('-v', '--verbose', action="count", help='verbose')
        # ------- mode: extract ----------------------------------------------------------------------------------------
        parser.add_argument('-i', '--input',
                            type=str, nargs='+', help='mode: extract. input folder')
        parser.add_argument('-o', '--output',
                            type=str, nargs='+',
                            help='mode: extract. output folder')
        # ------- mode: merge and place --------------------------------------------------------------------------------
        parser.add_argument('-t', '--targets',
                            type=str, nargs='+',
                            help='mode: merge and place. input target filenames')
        parser.add_argument('-n', '--name',
                            type=str, nargs=1,
                            help='mode: merge and place. output compound name')
        # ------- mode: place ------------------------------------------------------------------------------------------
        parser.add_argument('-s', '--smiles', type=str, nargs='+', help='mode: place. SMILE-String to be placed')
        return parser

    def verify(self):
        if self.args.mode == 'extract' and not (self.args.input and self.args.output):
            self.parser.error('mode extract requires --input <folder> and --output <folder>')
        elif self.args.mode == 'extract' and not self.args.targets:
            self.parser.error('mode extract requires --targets <file names>')
        elif self.args.mode == 'place' and not (self.args.targets and self.args.smiles):
            self.parser.error('mode place requires --targets <file names> and --smiles <string>')
        else:
            pass  # no issues.

    def route(self):
        self.set_verbosity(self.args.verbose)
        if self.args.mode == 'extract':
            self.extract(self.args.input, self.args.output)
        elif self.args.mode == 'merge':
            self.merge(self.args.targets)
        elif self.args.mode == 'place':
            self.place(self.args.targets, self.args.smiles)
        else:
            raise SyntaxError('Mode choice impossible. Check declaration')


# ======================================================================================================================

def main():
    FragmensteinParser()


if __name__ == '__main__':
    main()


import enum


class VictorMinMode(enum.Enum):
    """Victor mode for minimisation
    TODO this is not used yet
    """
    NULL = -1  #: no minimisation at all
    IGOR = 0  #: use Igor's minimisation (PyRosetta)
    IGOR_NODISK = 1  #: use Igor's minimisation without writing to disk
    FRITZ = 2  #: use Fritz's minimisation (openMM)



from ._victor_plonk import _VictorPlonk

class _VictorOverridables(_VictorPlonk):

    def post_params_step(self):
        """
        This method is intended for make inherited mods easier.
        :return:
        """
        pass

    def post_monster_step(self):
        """
        This method is intended for make inherited mods easier.
        :return:
        """
        pass

    def pose_mod_step(self):
        """
        This method is intended for make inherited mods easier.
        :return:
        """
        pass

    def pre_igor_step(self):
        """
        This method is intended for make inherited mods easier.
        :return:
        """
        pass

    def post_igor_step(self):
        """
        This method is intended for make inherited mods easier.
        :return:
        """
        pass


########################################################################################################################

__doc__ = \
    """
    See GitHub documentation
    """
__author__ = "Matteo Ferla. [Github](https://github.com/matteoferla)"
__email__ = "matteo.ferla@gmail.com"
__date__ = "2022 A.D."
__license__ = "MIT"
__citation__ = ""
from .version import __version__

########################################################################################################################

from warnings import warn
from . import legacy as _  # monkeypatches singledispatchmethod TypedDict Unpack on older Pythons

try:
    from .igor import Igor
except ImportError as err:
    warn(f'Igor (minimizer) unavailable —{err}.', category=ImportWarning)

try:
    from .victor import Victor
except ImportError as err:
    warn(f'Victor (pipeline) unavailable —{err}.', category=ImportWarning)

try:
    from .multivictor import MultiVictorPlacement
except ImportError as err:
    warn(f'Victor (pipeline) unavailable —{err}.', category=ImportWarning)

try:
    from .laboratory import Laboratory, binarize, unbinarize, place_input_validator
except ImportError as err:
    warn(f'Laboratory unavailable —{err}.', category=ImportWarning)

try:
    from .openmm import OpenVictor, Fritz
except ImportError as err:
    warn(f'OpenVictor / Fritz unavailable —{err}.', category=ImportWarning)

from .monster import Monster, MinizationOutcome
from .walton import Walton
from molecular_rectifier import Rectifier
from .m_rmsd import mRMSD
from .mpro import MProVictor
from .mpro import data as mpro_data
from .branding import divergent_colors, feijoa
from .display import display_mols, MolNGLWidget, patched_3Dmol_view, py3Dmol_monkey_patch, color_in, DISPLAYMODE
from .error import FragmensteinError, DistanceError, ShoddyCodeError, PoisonError
from .faux_victors import Wictor, Quicktor  # the rest are mostly experiments
from .settings import default_settings, cli_default_settings

if __name__ == '__main__':
    from .cli import main
    main()



from ._victor_store import _VictorStore
from rdkit import Chem
from rdkit.Chem import AllChem
from ..m_rmsd import mRMSD
from typing import Dict, Optional, Callable
from ..extraction_funs import combine_for_bondorder

class _VictorIgor(_VictorStore):

    def _fix_minimized(self, ligand: Optional[Chem.Mol]=None, add_dummy:bool=True) -> Chem.Mol:
        """
        PDBs are terrible for bond order etc. and Rosetta addes these based on atom types
        igor.mol_from_pose cannot use victor.params.mol because it's upstream
        hence this method which adds to it.

        :return:
        """
        if not add_dummy:
            self.journal.debug(f'{self.long_name} - making ligand only')
        else:
            self.journal.debug(f'{self.long_name} - making ligand w/ dummy (if present)')
        if ligand is None:  # normal route
            ligand = self.igor.mol_from_pose(add_dummy=add_dummy)
        new_ligand: Chem.Mol = combine_for_bondorder(self.params.mol, ligand)
        # self.journal.warning(f'{self.long_name} - Rosetta ring closure failed: +infinity kcal/mol penalty')
        # self.energy_score['bound']['total_score'] = float('inf')
        return new_ligand

    def quick_reanimate(self) -> float:
        """
        Correct small deviations from what the forcefield likes.
        Generally flattens buckled rings and that is it.
        Reanimate is normal.

        :return:
        """
        self.igor.coordinate_constraint = 10.
        self.igor.minimize(cycles=5, default_coord_constraint=False)
        self.energy_score = self.calculate_score()
        dG_bound = self.energy_score['bound']['total_score']
        dG_unbound = self.energy_score['unbound']['total_score']
        ddG = dG_bound - dG_unbound
        return ddG

    def reanimate(self) -> float:
        """
        Calls Igor recursively until the ddG is negative or zero.
        igor.minimize does a good job. this is just to get everything as a normal molecule

        :return: ddG (kcal/mol)
        """
        ddG = 999
        self.igor.coordinate_constraint = 0.
        # self.igor.fa_intra_rep = 0.02 # 4x
        # quick unconstrained minimisation to wiggle it out of nasty local minima
        self.igor.minimize(cycles=15, default_coord_constraint=False)
        self.igor.coordinate_constraint = 2
        self.igor.minimize(cycles=5, default_coord_constraint=False)
        self.igor.coordinate_constraint = 1
        while ddG > 0:
            self.journal.debug(f'{self.long_name} - Igor minimising')
            self.igor.minimize(default_coord_constraint=False)
            self.energy_score = self.calculate_score()
            dG_bound = self.energy_score['bound']['total_score']
            dG_unbound = self.energy_score['unbound']['total_score']
            ddG = dG_bound - dG_unbound
            if ddG > 0:
                self.igor.coordinate_constraint /= 2
                self.journal.debug(
                    f'{self.long_name} - coord_constraint lowered: {self.igor.coordinate_constraint}:  {ddG} kcal/mol.')
            if self.igor.coordinate_constraint == 0.:
                self.journal.warning(f'{self.long_name} - failed to minimise without constraints:  {ddG} kcal/mol.')
                break
            elif self.igor.coordinate_constraint < 0.005:
                self.igor.coordinate_constraint = 0.
        self.ddG = ddG
        return ddG

    def reanimate_n_store(self):
        self.reanimate()
        self._store_after_reanimation()

    def _store_after_reanimation(self):
        self.minimized_pdbblock = self.igor.pose2str()
        self.post_igor_step()  # empty overridable
        # extract_mol is a classmethod in _victor_utils: it could be an option but it's inputs differ
        # and there's a few diffent details.
        self.minimized_mol = self._fix_minimized()
        self.mrmsd = self._calculate_rmsd()
        self.journal.info(f'{self.long_name} - final score: {self.ddG} kcal/mol, RMSD: {self.mrmsd.mrmsd}.')
        self._checkpoint_charlie()
        self.journal.debug(f'{self.long_name} - Completed')

    def _calculate_rmsd(self):
        self.journal.debug(f'{self.long_name} - calculating mRMSD')
        return mRMSD.from_other_annotated_mols(self.minimized_mol, self.hits, self.monster.positioned_mol)

    def calculate_score(self):
        return {**self.igor.ligand_score(),
                'unbound': self.igor.detailed_scores(self.unbound_pose, 1)}

    @property
    def constrained_atoms(self) -> int:
        """
        Do note that the whole Origins list contains hydrogens. So do not divided by len!
        :return:
        """
        try:
            conn = sum([o != [] for o in self.monster.origin_from_mol(self.monster.positioned_mol)])
        except KeyboardInterrupt as err:
            raise err
        except Exception as err:
            self.journal.warning(f'{self.long_name} - {err.__class__.__name__}: {err}')
            conn = float('nan')
        return conn

    @property
    def unconstrained_heavy_atoms(self) -> int:
        try:
            origins = self.monster.origin_from_mol(self.monster.positioned_mol)
            unconn = sum([o == [] and atom.GetSymbol() != 'H' for o, atom in
                          zip(origins, self.monster.positioned_mol.GetAtoms())])
        except KeyboardInterrupt as err:
            raise err
        except Exception as err:
            self.journal.warning(f'{self.long_name} - {err.__class__.__name__}: {err}')
            unconn = float('nan')
        return unconn


import unittest
from typing import *

from fragmenstein import Victor, Igor, mpro_data
from fragmenstein.mpro import MProVictor


# ======================================================================================================================

class MProPlaceTester(unittest.TestCase):

    def setUp(self):
        Igor.init_pyrosetta()

    def untest_red_herring(self):  # without the test_ word this will not run.
        """
        To a **human** this looks easy. x0692 is a red herring and the other two make the molecule.
        As currently written this will fail.

        See red herring test notes in documentation.

        :return:
        """
        # PAU-UNI-52c0427f-1
        MProVictor.quick_reanimation = True
        victor = MProVictor.from_hit_codes(hit_codes=['x0692', 'x0305', 'x1249'])
        victor.place('CCNc1ncc(C#N)cc1CN1CCN(C(=O)C*)CC1', long_name='2_ACL')
        self.assertEqual(victor.error_msg, '', victor.error_msg)
        self.assertIsNotNone(victor.minimized_mol, 'Failed minimisation')
        msg = f'x1249 is the red herring, prediction: {victor.monster.unmatched} discarded, ' + \
              f'while x0305 and x0692 are the true inspirations. kept: {victor.monster.matched}'
        self.assertIn('x1249', victor.monster.unmatched, msg)
        self.assertIn('x0305', victor.monster.matched, msg)

        victor.make_pse()

    def untest_nasty(self):  # without the test_ word this will not run.
        """
        The human suggested a lot of novel groups.
        'x0540' is a really odd inspiration. Three atoms are conserved. the rest aren't.

        Like the red herring this is impossible.
        :return:
        """
        MProVictor.quick_reanimation = True
        victor = MProVictor.from_hit_codes(hit_codes=['x0434', 'x0540'])
        victor.place('*CCC(=O)N1CC(CCN(C(=O)Nc2c(C)ncc(C)c2CCN2CCOCC2)c2cc(C)ccn2)C1',
                     long_name='AGN-NEW-5f0-1_ACR1')
        self.assertEqual(str(victor.error_msg), '', str(victor.error_msg))
        self.assertIsNotNone(victor.minimized_mol, 'Failed minimisation')
        self.assertEqual(len(victor.monster.unmatched), 0,
                         f'Both were correct but {victor.monster.unmatched} was discarded')
        victor.make_pse()

    def test_incorrect(self):
        """
        This case has two hits that are not ideal. One, x0995, totally barney.
        These will be rejected.

        :return:
        """
        MProVictor.quick_reanimation = True
        victor = MProVictor.from_hit_codes(hit_codes='x0692,x0770,x0995'.split(','))
        victor.place('*C(=N)CN1CCN(Cc2ccc(-c3cc(CC)ncn3)c(F)c2)CC1',
                     long_name='BEN-VAN-c98-4')
        self.assertEqual(victor.error_msg, '', victor.error_msg)
        self.assertIsNotNone(victor.minimized_mol, 'Failed minimisation')
        # in unmatched, x0995 is the red herring.
        self.assertIn('x0995', victor.monster.unmatched, f'x0995 is the red herring')
        #victor.make_pse()

    def test_pentachromatic(self):
        """
        This hit fails to identify that the extra chloride comes from x1382.
        The human chose these.

        * the methylpyridine off x0107
        * the benzene-pyridine off x0434, but wanted an amide not a ureido
        * the link between the rings as amide x0678
        * x0995 is a red herring
        * benzene with a chroride from x1382
        """
        MProVictor.quick_reanimation = True
        # ,'x2646'
        Victor.monster_throw_on_discard = True
        victor = MProVictor.from_hit_codes(  # hit_codes=['x0107','x0434','x0678','x0995','x1382'],
            hit_codes=['x0107', 'x0434', 'x1382'])
        victor.place('Cc1ccncc1NC(=O)Cc1cccc(Cl)c1',
                     long_name='TRY-UNI-714a760b-6')
        self.assertEqual(victor.error_msg, '', victor.error_msg)
        self.assertIsNotNone(victor.minimized_mol, 'Failed minimisation')
        actual = mpro_data.get_mol('x2646')
        try:
            victor.make_pse(extra_mols=[actual])
        except ModuleNotFoundError:
            print('PyMOL is not installed. Cannot export result.')
        validation: Dict[str, float] = victor.validate(reference_mol=actual)
        rmsd = validation['reference2minimized_rmsd']
        self.assertLess(rmsd, 2, f'The RMSD is large...')
        self.assertIn('x1382', victor.monster.matched)
        self.assertNotIn('x0995', victor.monster.unmatched)  # red herring

if __name__ == '__main__':
    unittest.main()



import warnings
from typing import Optional, Dict, List, Tuple, Set, Unpack, Union  # noqa cf. .legacy monkeypatch

from rdkit import Chem
from rdkit.Chem import rdFMCS

from .._merge import _MonsterMerge
from ..mcs_mapping import SpecialCompareAtoms, IndexMap, ExtendedFMCSMode, transmute_FindMCS_parameters


class _MonsterMap(_MonsterMerge):
    def get_mcs_mappings(self,
                         hit: Chem.Mol,
                         followup: Chem.Mol,
                         min_mode_index: int = 0,
                         custom_map: Optional[Dict[str, Dict[int, int]]] = None
                         ) -> Tuple[List[Dict[int, int]], ExtendedFMCSMode]:
        """
        This is a curious method. It does a strict MCS match.
        And then it uses laxer searches and finds the case where a lax search includes the strict search.

        :param hit: query molecule
        :param followup: target/ref molecule
        :param min_mode_index: the lowest index to try (opt. speed reasons)
        :param custom_map: is the user defined hit name to list of tuples of hit and followup index pairs
        :return: mappings and mode
        """
        if not custom_map:
            # Issue 42:
            # Dict[str, IndexMap] = Dict[str, Sequence[Tuple[int, int]]]
            # not Dict[str, Dict[int, int]]
            custom_map: Dict[str, Dict[int, int]] = self.custom_map
        custom_map = self.fix_custom_map(custom_map)  # Issue 42
        hit_name: str = hit.GetProp('_Name')
        # -------------- Most strict mapping -------------------------------
        # run from strictest to laxest to find the strictest mapping that encompasses the custom map
        inverted_strict_i = 0
        for strict_i, mode in enumerate(reversed(self.matching_modes + [self.strict_matching_mode]), start=-1):
            # required for limiting the next iterator
            inverted_strict_i = len(self.matching_modes) - strict_i
            # this calls `self._get_atom_maps` does the MCS search
            strict_maps: List[IndexMap] = self._get_atom_maps(hit=hit,
                                                              followup=followup,
                                                              custom_map=custom_map,
                                                              **mode)
            # there is the possibility that the strict mapping does not allow the
            # provided custom_map
            # if so only the provided custom_map hits are used.
            if hit_name in self.custom_map:
                # wanted hit indices
                wanted: List[int] = self._get_required_indices_for_map(self.custom_map[hit_name])
                strict_maps: List[Dict[int, int]] = self._validate_vs_custom(strict_maps, wanted)
                if len(strict_maps) != 0:
                    # these maps are valid
                    break  # from the reverse loop...
        else:
            self.journal.warning('Provided mapping is very unfavourable... using that along for expanding the search')
            # unexpected pycharm warning as list({1:1}.items()) does give [[(1,1)]]
            # single choice list in this case
            strict_maps: List[Dict[int, int]] = [custom_map.get(hit_name, {}), ]
        self.journal.debug(f'`get_mcs_mappings` strict_maps for {hit_name}: {strict_maps}')
        # -------------- Expand mapping -------------------------------
        # go from laxest to strictest until one matches the strict form...
        for i, mode in enumerate(self.matching_modes):
            if i < min_mode_index:
                continue
            if i == inverted_strict_i:
                break
            # heme is 70 or so atoms & Any-Any matching gets stuck. So guestimate to avoid that:
            if hit.GetNumAtoms() > (50 + i * 10) or followup.GetNumAtoms() > (50 + i * 10):
                continue

            lax: List[IndexMap] = []
            for strict_map in strict_maps:  #: Dict[int, int]
                # `_get_atom_maps` does the MCS search constrained by `expanded_custom_map`
                expanded_custom_map: Dict[str, Dict[int, int]] = self.expand_custom_map(custom_map,
                                                                                        {hit_name: strict_map})
                lax.extend(self._get_atom_maps(hit=hit,
                                               followup=followup,
                                               custom_map=expanded_custom_map,
                                               **mode)
                           )
            if len(lax) == 0:
                continue
            else:
                return [dict(n) for n in lax], mode
        # The strict will have to do.
        return [dict(n) for n in strict_maps], self.strict_matching_mode  # tuple to dict

    def get_mcs_mapping(self, hit, followup, min_mode_index: int = 0) -> Tuple[Dict[int, int], dict]:
        """
        This is a weird method. It does a strict MCS match.
        And then it uses laxer searches and finds the case where a lax search includes the strict search.

        :param hit: query molecule
        :param followup: target/ref molecule
        :param min_mode_index: the lowest index to try (opt. speed reasons)
        :return: mapping and mode
        """
        ms, mode = self.get_mcs_mappings(hit, followup, min_mode_index)
        return ms[0], mode

    def _get_atom_maps(self,
                       hit: Chem.Mol,
                       followup: Chem.Mol,
                       custom_map: Optional[Dict[str, IndexMap]]=None,
                       **mode: Unpack[ExtendedFMCSMode]) -> List[Dict[int, int]]:

        """
        The ``mode`` are FindMCS arguments, but this transmutes them into parameter scheme

        The old method is now ``_get_atom_maps_OLD``.
        """
        if custom_map is None:
            custom_map: Dict[str, Dict[int, int]] = self.custom_map
        custom_map = self.fix_custom_map(custom_map)  # Issue 42
        parameters: rdFMCS.MCSParameters = transmute_FindMCS_parameters(**mode)
        # this looks odd, because the default parameters.AtomTyper is a atomcompare enum
        # and can be overridden by a callable class instance (of MCSAtomCompare)
        parameters.AtomTyper = SpecialCompareAtoms(comparison=parameters.AtomTyper,
                                                   custom_map=custom_map)
        res: rdFMCS.MCSResult = rdFMCS.FindMCS([hit, followup], parameters)
        matches: List[Dict[int, int]] = parameters.AtomTyper.get_valid_matches(parameters.AtomCompareParameters,
                                                                         common=Chem.MolFromSmarts(res.smartsString),
                                                                         hit=hit,
                                                                         followup=followup
                                                                         )
        return [dict(m) for m in matches]

    def _get_atom_maps_OLD(self, molA, molB, **mode: Unpack[ExtendedFMCSMode]) -> Set[Tuple[Tuple[int, int]]]:
        """
        The ``mode`` are FindMCS arguments.
        """
        warnings.warn('`_get_atom_maps_OLD` is no longer used.', category=DeprecationWarning)
        mcs = rdFMCS.FindMCS([molA, molB], **mode)
        common = Chem.MolFromSmarts(mcs.smartsString)
        matches = []
        # prevent a dummy to match a non-dummy, which can happen when the mode is super lax.
        is_dummy = lambda mol, at: mol.GetAtomWithIdx(at).GetSymbol() == '*'  # noqa: E731 is stupid

        def all_bar_dummy(Aat, Bat) -> bool:
            """it is okay to match a dummy with dummy only"""
            a_dummy: bool = is_dummy(molA, Aat)
            b_dummy: bool = is_dummy(molB, Bat)
            # xor
            return (a_dummy and b_dummy) or not (a_dummy or b_dummy)

        # prevent matching hydrogens
        is_hydrogen = lambda mol, at: mol.GetAtomWithIdx(at).GetSymbol() == 'H'  # noqa: E731 is stupid
        for molA_match in molA.GetSubstructMatches(common, uniquify=False):
            for molB_match in molB.GetSubstructMatches(common, uniquify=False):
                matches.append([(int(molA_at), int(molB_at)) for molA_at, molB_at in zip(molA_match, molB_match) if
                                all_bar_dummy(molA_at, molB_at) and
                                not is_hydrogen(molA, molA_at) and
                                not is_hydrogen(molB, molB_at)])
        # you can map two toluenes 4 ways, but two are repeats.
        matches = set([tuple(sorted(m, key=lambda i: i[0])) for m in matches])
        return matches

    def _get_atom_map(self, molA, molB, **mode) -> List[Tuple[int, int]]:
        return self._get_atom_maps(molA, molB, **mode)[0]

    def expand_custom_map(self,
                          custom_map: Dict[str, Dict[int, int]],
                          addend: Dict[str, Dict[int, int]]) \
            -> Dict[str, Dict[int, int]]:
        # not a IndexMap
        new_map = custom_map.copy()
        for k, add_mapping in addend.items():
            if k not in new_map:
                new_map[k] = dict(add_mapping)
            else:
                # custom_map gets priority
                new_map[k] = {**dict(add_mapping), **dict(custom_map[k])}
        return new_map

    def _validate_vs_custom(self, maps: List[Dict[int, int]], wanted_idx: List[int]) -> List[Dict[int, int]]:
        """return only the IndexMaps with all wanted_idx (wanted in the hit)
        This is not part of ``SpecialCompareAtoms.get_valid_matches``
        because there is a difference between must match and must be present as discussed in that method.
        """
        get_set = lambda mapping: set(mapping.keys()) if isinstance(mapping, dict) else {i for i, j in mapping}
        return [mapping for mapping in maps if len(set(wanted_idx) - get_set(mapping)) == 0]

    def _get_required_indices_for_map(self, custom_hit_map: Dict[int, int]) -> List[int]:
        """
        Returns the hit indices that must be present in the map.
        """
        return [h for h, f in custom_hit_map.items() if h >= 0 and f >= 0]
